/****************************************************************************
** Generated QML type registration code
**
** WARNING! All changes made in this file will be lost!
*****************************************************************************/

#include <QtQml/qqml.h>
#include <QtQml/qqmlmoduleregistration.h>

#if __has_include(<private/qqmlwebchannelattached_p.h>)
#  include <private/qqmlwebchannelattached_p.h>
#endif
#if __has_include(<qqmlwebchannel.h>)
#  include <qqmlwebchannel.h>
#endif


#if !defined(QT_STATIC)
#define Q_QMLTYPE_EXPORT Q_DECL_EXPORT
#else
#define Q_QMLTYPE_EXPORT
#endif
Q_QMLTYPE_EXPORT void qml_register_types_QtWebChannel()
{
    qmlRegisterModule("QtWebChannel", 1, 0);
    qmlRegisterModule("QtWebChannel", 1, 254);
    qmlRegisterModule("QtWebChannel", 6, 0);
    QT_WARNING_PUSH QT_WARNING_DISABLE_DEPRECATED
    qmlRegisterTypesAndRevisions<QQmlWebChannel>("QtWebChannel", 6);
    qmlRegisterTypesAndRevisions<QQmlWebChannelAttached>("QtWebChannel", 6);
    QMetaType::fromType<QWebChannel *>().id();
    QT_WARNING_POP
    qmlRegisterModule("QtWebChannel", 6, 12);
}

static const QQmlModuleRegistration qtWebChannelRegistration("QtWebChannel", qml_register_types_QtWebChannel);
